/*This file contains all of the classes used by the game engine, as well as the loadTexture_Ipl function*/

int loadTexture_Ipl(IplImage *image, GLuint *text) {
  /*modified code from http://blog.damiles.com/?p=9*/

  if (image==NULL) return -1;

  glGenTextures(1, text);

  glBindTexture( GL_TEXTURE_2D, *text ); //bind the texture to its array
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);

  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

  glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, image->width, image->height,0, GL_BGR_EXT, GL_UNSIGNED_BYTE, image->imageData);

  return 0;
}


class visual{
  /*The visual class is a base class for all other classes that use images, such as background objects
    and sprites. It contains methods for specifying the image, the location (center) of the image in OpenGL
    coordinates, orientation and scale, and drawing the image.*/

public:
  GLuint tex; 			//texture
  int pixwidth, pixheight;
  float width, height;	//image width and height in OpenGL coordinates
  float xscale, yscale;	//scaling factors
  float x, y, z, theta;	//OpenGL coords of image center + rotation
  int visible;

  visual() {
    xscale = 1;
    yscale = 1;
    visible = 1;
  }

  int setImage(const char string[]){
    /*set the image*/
    IplImage *image = NULL;
    image = cvLoadImage(string);
    if(!image)
      return 1;	//something went wrong
    pixwidth = image->width;
    pixheight = image->height;
    width = pixwidth*widthscale;		//in OpenGL coordinates
    height = pixheight*heightscale;	//in OpenGL coordinates
    loadTexture_Ipl(image, &tex);
    cvReleaseImage(&image);
    return 0;
  }

  int clearImage(){
    glDeleteTextures(1, &tex);
  }	

  int setXYZ(float xx, float yy, float zz){
    /*set the x,y,z coords of image center*/
    x = xx; y = yy; z = zz;
    return 0;
  }

  int setTheta(float t){
    /*set image rotation - degrees counterclockwise*/
    theta = t;
    return 0;
  }


  int setVisible(int v){
    /*set visibility status of object*/
    visible = v;
    return 0;
  }

  int setScale(float xs, float ys){
    /*set scaling coefficients*/
    xscale = xs;
    yscale = ys;
    return 0;
  }

  virtual int draw(){
    /*perform OpenGL calls to draw image - virtual to allow masking operations in sprite class*/
    float ww, hh;

    ww = width*xscale*0.5;
    hh = height*yscale*0.5;

    if(!visible){	//we're coastin' for a ghostin'
      return 0;	//so don't do anything
    }

    glPushMatrix();
    glTranslatef(x,y,0);
    glRotatef(theta,0,0,1);
    glBindTexture(GL_TEXTURE_2D, tex);
    glBegin(GL_QUADS);
    glTexCoord2f(0.0, 1.0); glVertex3f(-ww, -hh, z);	//bottom left
    glTexCoord2f(1.0, 1.0); glVertex3f(ww, -hh, z);		//bottom right
    glTexCoord2f(1.0, 0.0); glVertex3f(ww, hh, z);		//top right
    glTexCoord2f(0.0, 0.0); glVertex3f(-ww, hh, z);		//top left
    glEnd();
    glPopMatrix();
    return 0;
  }

};

class background: public visual{
  /*visual with a second b/w image that is used to determine boundaries through which sprites don't pass*/
public:
  IplImage *bndimage;

  int setBoundaryImage(const char bndstring[]){
    bndimage = cvLoadImage(bndstring);	
    if(!bndimage)
      return 1;	//error
    return 0;
  }
};


class maskvisual: public visual{
  /*modified visual class that works with masks to create "odd" shaped visuals w/ help of blending*/ 
public:
  GLuint masktex;

  int setMask(const char string[]){
    /*set mask image - basically same code as for setImage method in visual class*/
    IplImage *image = NULL;
    image = cvLoadImage(string);
    if(!image)
      return 1;	//something went wrong
    loadTexture_Ipl(image, &masktex);
    cvReleaseImage(&image);
    return 0;	
  }

  int draw(){
    /*differs slightly from virtual method in visual class*/

    float ww, hh;

    ww = width*xscale*0.5;
    hh = height*yscale*0.5;

    if(!visible){
      return 0;
    }

    glPushMatrix();
    glTranslatef(x,y,0);
    glRotatef(theta,0,0,1);

    glEnable(GL_BLEND);
    glDisable(GL_DEPTH_TEST);

    glBlendFunc(GL_ONE_MINUS_SRC_COLOR,GL_SRC_COLOR);

    glBindTexture(GL_TEXTURE_2D, masktex);
    glBegin(GL_QUADS);
    glTexCoord2f(0.0, 1.0); glVertex3f(-ww, -hh, z);	//bottom left
    glTexCoord2f(1.0, 1.0); glVertex3f(ww, -hh, z);		//bottom right
    glTexCoord2f(1.0, 0.0); glVertex3f(ww, hh, z);		//top right
    glTexCoord2f(0.0, 0.0); glVertex3f(-ww, hh, z);		//top left
    glEnd();

    glBlendFunc(GL_SRC_COLOR,GL_ONE);

    glBindTexture(GL_TEXTURE_2D, tex);
    glBegin(GL_QUADS);
    glTexCoord2f(0.0, 1.0); glVertex3f(-ww, -hh, z);	//bottom left
    glTexCoord2f(1.0, 1.0); glVertex3f(ww, -hh, z);	//bottom right
    glTexCoord2f(1.0, 0.0); glVertex3f(ww, hh, z);	//top right
    glTexCoord2f(0.0, 0.0); glVertex3f(-ww, hh, z);	//top left
    glEnd();

    glPopMatrix();

    glDisable(GL_BLEND);
    glEnable(GL_DEPTH_TEST);

    return 0;
  }
};

class sprite: public maskvisual{
  /*extension of maskvisual class that adds hitbox functionality and virtual onHit method.
    Possible future extension: record data about what hit the sprite, not just that it was hit -
    could do this by assigning each sprite or sprite-daughter class a unique int classID attribute
    and use the classID in a switch statement in the onHit method - might not need this functionality,
    though*/
public:
  /*hit box corner positions relative to sprite's center
    naming example: hbt = "HitBox Top (distance from center)" */
  float hbt, hbb, hbl, hbr;
  int facing;	//0 up, 1 right, 2 down, 3 left
  int inactive;	//0=respond to onHit, 1=do nothing for onHit
  sprite* othersprite;	//pointer to the last sprite that this sprite was hit by (or NULL)

  int setHitBox(float t, float b, float l, float r){
    /*have to set boundaries before any calls to isHitBy otherwise it is impossible to detect hit*/
    hbt = t; hbb = b; hbl = l; hbr = r;	//hbt = "hit box top" boundary
    return 0;
  }

  int isHitBy(sprite *sp){
    /*only interactive things (sprites or any daughter classes) should be able to hit a sprite
      (or any daughter class). sp is the sprite we're checking to see if we're hit by.
      Ex. character.isHitBy(enemy) would check to see if sprite character is hit by sprite enemy.*/ 
    float spx = sp->x, spy = sp->y, spt = sp->hbt, spb = sp->hbb, spl = sp->hbl, spr = sp->hbr,
      spxscale = sp->xscale, spyscale = sp->yscale;

    if(spx+spr*spxscale > x-hbl*xscale && spx-spl*spxscale < x+hbr*xscale && 
       spy+spt*spyscale > y-hbb*yscale && spy-spb*spyscale < y+hbt*yscale){
      othersprite = sp;
      return 1;
    }
    else{
      othersprite = NULL;
      return 0;
    }
  }

  int setInactive(int in){
    inactive = in;
    return 0;
  }

  virtual int onHit(){
    /*print out hit information as default behavior. this will be overridden by the onHit() method
      in actual derived classes*/
    if(inactive){
      return 0;
    }
    printf("sprite %lu was hit\n", (long unsigned int)this);
    return 0;
  }

  int setFacing(int f){
    /*self explanatory, really...*/
    facing = f;
    return 0;
  }

  int trySetXYZ(float xx, float yy, float zz, background* bg){
    /*try to set xyz. fails if there is a background obstacle in the way*/
    int pixelx, pixely;
    int bndval;
    pixelx = xx/widthscale;
    pixelx += bg->pixwidth*0.5*bg->xscale;
    pixely = bg->pixheight*0.5*bg->yscale;
    pixely -= yy/heightscale;

    pixelx /= bg->xscale;
    pixely /= bg->yscale;

    //check obstacle
    //printf("\npixel (x,y) = (%d,%d)\n",pixelx,pixely);
    //printf("\nval = %d\n", (uchar)*(bg->bndimage->imageData + pixely*bg->bndimage->widthStep +pixelx*bg->bndimage->nChannels));
    bndval = (uchar)*(bg->bndimage->imageData + pixely*bg->bndimage->widthStep +pixelx*bg->bndimage->nChannels);
    if(bndval){
      x = xx;
      y=yy;
      return 0;	//set succeeded
    }

    return 1;	//set failed
  }
};


class obstacle: public sprite{
  /*interacts with sprites, and so must also be a sprite according to my arbitrary rules. OBEY OBEY OBEY.
    Really, what separates this class from sprite is a specific onHit() function.*/
public:
	
  virtual int onHit(){
    /*block any sprite that hits it, so the onHit action is registered in obstacle, not whatever hits obstacle (saves code).*/	

    if(inactive){
      return 0;
    }
		
    switch(othersprite->facing){
    case 0:{
      othersprite->y -= move_amnt;
      break;
    }

    case 1:{
      othersprite->x -= move_amnt;
      break;
    }

    case 2:{
      othersprite->y += move_amnt;
      break;
    }

    case 3:{
      othersprite->x += move_amnt;
      break;
    }

    default: break;
    }

    /*spin to show your anger at being hit!*/
    theta += 15;
  }	
};


class transporter: public obstacle{
  /*same as obstacle, but use onHit() to change currentscene game state*/
public:

  int nextscene;	//what scene will we transport to?
		
  int setNextScene(int ns){
    nextscene = ns;
    return 0;
  }

  int onHit(){
    /*block any sprite that hits it, so the onHit action is registered in obstacle, not whatever hits obstacle (saves code).*/	

    if(inactive){
      return 0;
    }
			
    currentscene = nextscene;
    newscene = 1;	//trigger initialization routine for next scene

    return 0;
  }
};

